function [R,T,Xc,best_solution,opt]=efficient_pnp_gauss(Xw,U,A)

%clear all; close all; load efficient_pnp_gauss;

% EFFICIENT_PNP_GAUSS Main Function to solve the PnP problem 
%       as described in:
%
%       Francesc Moreno-Noguer, Vincent Lepetit, Pascal Fua.
%       Accurate Non-Iterative O(n) Solution to the PnP Problem. 
%       In Proceedings of ICCV, 2007. 
%
%       Note: In this version of the software we perform a final
%       optimization using Gauss-Newton,which is not described in the
%       paper.
%
%       x3d_h: homogeneous coordinates of the points in world reference
%       x2d_h: homogeneous position of the points in the image plane
%       A: intrincic camera parameters
%       R: Rotation of the camera system wrt world reference
%       T: Translation of the camera system wrt world reference
%       Xc: Position of the points in the camera reference
%       best solution: dimension of the kernel for the best solution
%                     (before applying Gauss Newton).
%       opt: some parameters of the optimization process
%
% Copyright (C) <2007>  <Francesc Moreno-Noguer, Vincent Lepetit, Pascal Fua>
% 
% This program is free software: you can redistribute it and/or modify
% it under the terms of the version 3 of the GNU General Public License
% as published by the Free Software Foundation.
% 
% This program is distributed in the hope that it will be useful, but
% WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
% General Public License for more details.       
% You should have received a copy of the GNU General Public License
% along with this program. If not, see <http://www.gnu.org/licenses/>.
%
% Francesc Moreno-Noguer, CVLab-EPFL, October 2007.
% fmorenoguer@gmail.com, http://cvlab.epfl.ch/~fmoreno/ 



% Xw=x3d_h(:,1:3);
% U=x2d_h(:,1:2);

THRESHOLD_REPROJECTION_ERROR=3000;%error in degrees of the basis formed by the control points. 
%If we have a larger error, we will compute the solution using a larger
%number of vectors in the kernel

%define control points in a world coordinate system (centered on the 3d
%points centroid)
Cw=define_control_points();

%compute alphas (linear combination of the control points to represent the 3d
%points)
Alph=compute_alphas(Xw,Cw);

%Compute M
M=compute_M_ver2(U,Alph,A);

%Compute kernel M
Km=kernel_noise(M,4); %in matlab we have directly the funcion km=null(M);
   


%1.-Solve assuming dim(ker(M))=1. X=[Km_end];------------------------------
dim_kerM=1;
X1=Km(:,end);

[Cc,Xc,sc]=compute_norm_sign_scaling_factor(X1,Cw,Alph,Xw);

[R,T]=getrotT(Xw,Xc);  %solve exterior orientation
err(1)=reprojection_error_usingRT(Xw,U,R,T,A);

sol(1).Xc=Xc;
sol(1).Cc=Cc;
sol(1).R=R;
sol(1).T=T;
sol(1).error=err(1);
sol(1).betas=[1];
sol(1).sc=sc;
sol(1).Kernel=X1;



%2.-Solve assuming dim(ker(M))=2------------------------------------------
Km1=Km(:,end-1);
Km2=Km(:,end);

%control points distance constraint
D=compute_constraint_distance_2param_6eq_3unk(Km1,Km2);
dsq=define_distances_btw_control_points();
betas_=inv(D'*D)*D'*dsq;
beta1=sqrt(abs(betas_(1)));
beta2=sqrt(abs(betas_(3)))*sign(betas_(2))*sign(betas_(1));
X2=beta1*Km1+beta2*Km2;

[Cc,Xc,sc]=compute_norm_sign_scaling_factor(X2,Cw,Alph,Xw);

[R,T]=getrotT(Xw,Xc);  %solve exterior orientation
err(2)=reprojection_error_usingRT(Xw,U,R,T,A);

sol(2).Xc=Xc;
sol(2).Cc=Cc;
sol(2).R=R;
sol(2).T=T;
sol(2).error=err(2);
sol(2).betas=[beta1,beta2];
sol(2).sc=sc;
sol(2).Kernel=[Km1,Km2];





%3.-Solve assuming dim(ker(M))=3------------------------------------------
if min(err)>THRESHOLD_REPROJECTION_ERROR %just compute if we do not have good solution in the previus cases

    Km1=Km(:,end-2);
    Km2=Km(:,end-1);
    Km3=Km(:,end);

    %control points distance constraint
    D=compute_constraint_distance_3param_6eq_6unk(Km1,Km2,Km3);
    dsq=define_distances_btw_control_points();
    betas_=inv(D)*dsq;
    beta1=sqrt(abs(betas_(1)));
    beta2=sqrt(abs(betas_(4)))*sign(betas_(2))*sign(betas_(1));
    beta3=sqrt(abs(betas_(6)))*sign(betas_(3))*sign(betas_(1));

    X3=beta1*Km1+beta2*Km2+beta3*Km3;

    [Cc,Xc,sc]=compute_norm_sign_scaling_factor(X3,Cw,Alph,Xw);
  
    [R,T]=getrotT(Xw,Xc);  %solve exterior orientation
    err(3)=reprojection_error_usingRT(Xw,U,R,T,A);

    sol(3).Xc=Xc;
    sol(3).Cc=Cc;
    sol(3).R=R;
    sol(3).T=T;
    sol(3).error=err(3);
    sol(3).betas=[beta1,beta2,beta3];
    sol(3).sc=sc;
    sol(3).Kernel=[Km1,Km2,Km3];

end




%4.-Solve assuming dim(ker(M))=4------------------------------------------
if min(err)>THRESHOLD_REPROJECTION_ERROR %just compute if we do not have good solution in the previus cases
    Km1=Km(:,end-3);
    Km2=Km(:,end-2);
    Km3=Km(:,end-1);
    Km4=Km(:,end);


    D=compute_constraint_distance_orthog_4param_9eq_10unk(Km1,Km2,Km3,Km4);
    dsq=define_distances_btw_control_points();
    lastcolumn=[-dsq',0,0,0]';
    D_=[D,lastcolumn];
    Kd=null(D_);

    P=compute_permutation_constraint4(Kd);
    lambdas_=kernel_noise(P,1);
    lambda(1)=sqrt(abs(lambdas_(1)));
    lambda(2)=sqrt(abs(lambdas_(6)))*sign(lambdas_(2))*sign(lambdas_(1));
    lambda(3)=sqrt(abs(lambdas_(10)))*sign(lambdas_(3))*sign(lambdas_(1));
    lambda(4)=sqrt(abs(lambdas_(13)))*sign(lambdas_(4))*sign(lambdas_(1));
    lambda(5)=sqrt(abs(lambdas_(15)))*sign(lambdas_(5))*sign(lambdas_(1));

    betass_=lambda(1)*Kd(:,1)+lambda(2)*Kd(:,2)+lambda(3)*Kd(:,3)+lambda(4)*Kd(:,4)+lambda(5)*Kd(:,5);
    beta1=sqrt(abs(betass_(1)));
    beta2=sqrt(abs(betass_(5)))*sign(betass_(2));
    beta3=sqrt(abs(betass_(8)))*sign(betass_(3));
    beta4=sqrt(abs(betass_(10)))*sign(betass_(4));
    X4=beta1*Km1+beta2*Km2+beta3*Km3+beta4*Km4;

    [Cc,Xc,sc]=compute_norm_sign_scaling_factor(X4,Cw,Alph,Xw);
    
    [R,T]=getrotT(Xw,Xc);  %solve exterior orientation
    err(4)=reprojection_error_usingRT(Xw,U,R,T,A);

    sol(4).Xc=Xc;
    sol(4).Cc=Cc;
    sol(4).R=R;
    sol(4).T=T;
    sol(4).error=err(4);
    sol(4).betas=[beta1,beta2,beta3,beta4];
    sol(4).sc=sc;
    sol(4).Kernel=[Km1,Km2,Km3,Km4];
end




%5.-Gauss Newton Optimization------------------------------------------------------ 
[min_err,best_solution]=min(err);
Xc=sol(best_solution).Xc;
R=sol(best_solution).R;
T=sol(best_solution).T;
Betas=sol(best_solution).betas;
sc=sol(best_solution).sc;
Kernel=sol(best_solution).Kernel;
 
if best_solution==1
    Betas=[0,0,0,Betas];
elseif best_solution==2
    Betas=[0,0,Betas];
elseif best_solution==3
    Betas=[0,Betas];
end

Km1=Km(:,end-3);
Km2=Km(:,end-2);
Km3=Km(:,end-1);
Km4=Km(:,end);
Kernel=[Km1,Km2,Km3,Km4];


%refine the solution iterating over the betas
Beta0=Betas/sc;
[Xc_opt,R_opt,T_opt,err_opt,iter]=optimize_betas_gauss_newton(Kernel,Cw,Beta0,Alph,Xw,U,A);

%Just update R,T,Xc if Gauss Newton improves results (which is almost
%always)
if err_opt<min_err    
    R=R_opt;
    T=T_opt;
    Xc=Xc_opt;
end

opt.Beta0=Beta0;
opt.Kernel=Kernel;
opt.iter=iter;




%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function [R, T]=getrotT(wpts,cpts)
  
% This routine solves the exterior orientation problem for a point cloud
%  given in both camera and world coordinates. 
  
% wpts = 3D points in arbitrary reference frame
% cpts = 3D points in camera reference frame
  
n=size(wpts,1);
M=zeros(3);

ccent=mean(cpts);
wcent=mean(wpts);

for i=1:3
  cpts(:,i)=cpts(:,i)-ccent(i)*ones(n,1);
  wpts(:,i)=wpts(:,i)-wcent(i)*ones(n,1);
end
for i=1:n
   M=M+cpts(i,:)'*wpts(i,:);
end
[U S V]=svd(M);
R=U*V';
if det(R)<0
  R=-R;
end
T=ccent'-R*wcent';
% 


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function [err,Urep]=reprojection_error_usingRT(Xw,U,R,T,A)

%clear all; close all; load reprojection_error_usingRT;
n=size(Xw,1);

P=A*[R,T];
Xw_h=[Xw,ones(n,1)];
Urep_=(P*Xw_h')';

%project reference points into the image plane
Urep=zeros(n,2);
Urep(:,1)=Urep_(:,1)./Urep_(:,3);
Urep(:,2)=Urep_(:,2)./Urep_(:,3);


%reprojection error
err_=sqrt((U(:,1)-Urep(:,1)).^2+(U(:,2)-Urep(:,2)).^2);
err=sum(err_)/n;


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function Cw=define_control_points()

Cw=[1 0 0;
    0 1 0;
    0 0 1;
    0 0 0];

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function Alph=compute_alphas(Xw,Cw)

n=size(Xw,1); %number of 3d points

%generate auxiliar matrix to compute alphas
C=[Cw';ones(1,4)];
X=[Xw';ones(1,n)];
Alph_=inv(C)*X;



Alph=Alph_';

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function M=compute_M_ver2(U,Alph,A)


n=size(Alph,1); %number of 3d points

fu=A(1,1);
fv=A(2,2);
u0=A(1,3);
v0=A(2,3);

nrows_M=2*n;
ncols_M=12;
M=zeros(nrows_M,ncols_M);



for i=1:n
    a1=Alph(i,1);
    a2=Alph(i,2);
    a3=Alph(i,3);
    a4=Alph(i,4);
    
    ui=U(i,1);
    vi=U(i,2);
    
    %generate submatrix M
    M_=[a1*fu, 0, a1*(u0-ui), a2*fu, 0, a2*(u0-ui), a3*fu, 0, a3*(u0-ui), a4*fu, 0, a4*(u0-ui);
        0, a1*fv, a1*(v0-vi), 0, a2*fv, a2*(v0-vi), 0, a3*fv, a3*(v0-vi), 0, a4*fv, a4*(v0-vi)];
    
    
    %put M_ in the whole matrix
    row_ini=i*2-1;
    row_end=i*2;
        
    M(row_ini:row_end,:)=M_;
     
end
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function K=kernel_noise(M,dimker)

MtM=M'*M;
[V,S]=eig(MtM);

K=V(:,dimker:-1:1);


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function [Cc,Xc,sc]=compute_norm_sign_scaling_factor(X1,Cw,Alph,Xw)
 
n=size(Xw,1); %number of data points

%Km will be a scaled solution. In order to find the scale parameter we
%impose distance constraints between the reference points

%scaled position of the control points in camera coordinates
Cc_=zeros(4,3);
for i=1:4
    Cc_(i,:)=X1(3*i-2:3*i);
end

%position of reference points in camera coordinates
Xc_=Alph*Cc_;

%compute distances in world coordinates w.r.t. the centroid
centr_w=mean(Xw);
centroid_w=repmat(centr_w,[n,1]);
tmp1=Xw-centroid_w;
dist_w=sqrt(sum(tmp1.^2,2));

%compute distances in camera coordinates w.r.t. the centroid
centr_c=mean(Xc_);
centroid_c=repmat(centr_c,[n,1]);
tmp2=Xc_-centroid_c;
dist_c=sqrt(sum(tmp2.^2,2));
 
%least squares solution for the scale factor
sc=1/(inv(dist_c'*dist_c)*dist_c'*dist_w);

%scale position of the control points
Cc=Cc_/sc;

%rescaled position of the reference points
Xc=Alph*Cc;

%change the sign if necessary. z negative is no possible in camera
%coordinates
neg_z=find(Xc(:,3)<0);
if size(neg_z,1)>=1
    sc=-sc;
    Xc=Xc*(-1);
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function P=compute_constraint_distance_2param_6eq_3unk(m1,m2)

%redefine variables name, for compatibility with maple
m1_1=m1(1); 
m1_2=m1(2); 
m1_3=m1(3); 
m1_4=m1(4); 
m1_5=m1(5); 
m1_6=m1(6);
m1_7=m1(7); 
m1_8=m1(8); 
m1_9=m1(9); 
m1_10=m1(10); 
m1_11=m1(11); 
m1_12=m1(12);

m2_1=m2(1); 
m2_2=m2(2); 
m2_3=m2(3); 
m2_4=m2(4); 
m2_5=m2(5); 
m2_6=m2(6);
m2_7=m2(7); 
m2_8=m2(8); 
m2_9=m2(9); 
m2_10=m2(10); 
m2_11=m2(11); 
m2_12=m2(12);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

t7 = (m1_6 ^ 2);
t8 = (m1_4 ^ 2);
t9 = (m1_1 ^ 2);
t10 = (m1_5 ^ 2);
t11 = (m1_2 ^ 2);
t12 = (m1_3 ^ 2);
t17 = m1_4 * m2_4;
t18 = m1_1 * m2_1;
t19 = m1_5 * m2_5;
t22 = m1_2 * m2_2;
t23 = m1_6 * m2_6;
t25 = m1_3 * m2_3;
t26 = (-m2_6 * m1_3 - m1_4 * m2_1 - m2_4 * m1_1 + t17 + t18 + t19 - m1_5 * m2_2 - m2_5 * m1_2 + t22 + t23 - m1_6 * m2_3 + t25);
t29 = (m2_3 ^ 2);
t34 = (m2_4 ^ 2);
t35 = (m2_1 ^ 2);
t36 = (m2_5 ^ 2);
t37 = (m2_2 ^ 2);
t38 = (m2_6 ^ 2);
t44 = (m1_7 ^ 2);
t45 = (m1_8 ^ 2);
t46 = (m1_9 ^ 2);
t55 = m1_8 * m2_8;
t56 = m1_9 * m2_9;
t58 = m1_7 * m2_7;
t59 = (-m1_9 * m2_3 - m2_8 * m1_2 - m2_9 * m1_3 - m1_7 * m2_1 - m2_7 * m1_1 + t55 + t22 + t56 + t18 - m1_8 * m2_2 + t25 + t58);
t64 = (m2_8 ^ 2);
t65 = (m2_9 ^ 2);
t68 = (m2_7 ^ 2);
t72 = (m1_11 ^ 2);
t73 = (m1_12 ^ 2);
t74 = (m1_10 ^ 2);
t85 = m1_10 * m2_10;
t86 = m1_11 * m2_11;
t88 = m1_12 * m2_12;
t89 = (-m1_10 * m2_1 - m2_10 * m1_1 - m1_12 * m2_3 - m2_11 * m1_2 - m1_11 * m2_2 + t18 + t22 + t25 + t85 + t86 - m2_12 * m1_3 + t88);
t92 = (m2_11 ^ 2);
t95 = (m2_12 ^ 2);
t96 = (m2_10 ^ 2);
t113 = (-m1_9 * m2_6 - m2_9 * m1_6 + t55 + t23 + t17 + t56 + t58 - m1_7 * m2_4 - m2_7 * m1_4 - m1_8 * m2_5 - m2_8 * m1_5 + t19);
t134 = (-m1_10 * m2_4 - m2_10 * m1_4 + t88 + t23 + t17 + t85 + t86 - m1_11 * m2_5 - m2_11 * m1_5 - m1_12 * m2_6 - m2_12 * m1_6 + t19);
t155 = (t58 + t88 - m2_10 * m1_7 - m2_11 * m1_8 + t56 - m1_10 * m2_7 + t55 + t85 + t86 - m1_12 * m2_9 - m2_12 * m1_9 - m1_11 * m2_8);
P(1,1) = -2 * m1_4 * m1_1 - 2 * m1_5 * m1_2 - 2 * m1_6 * m1_3 + t7 + t8 + t9 + t10 + t11 + t12;
P(1,2) = 2 * t26;
P(1,3) = -2 * m2_6 * m2_3 + t29 - 2 * m2_4 * m2_1 - 2 * m2_5 * m2_2 + t34 + t35 + t36 + t37 + t38;
P(2,1) = -2 * m1_7 * m1_1 + t12 - 2 * m1_9 * m1_3 + t44 + t45 + t46 - 2 * m1_8 * m1_2 + t9 + t11;
P(2,2) = 2 * t59;
P(2,3) = -2 * m2_8 * m2_2 - 2 * m2_9 * m2_3 + t64 + t65 - 2 * m2_7 * m2_1 + t29 + t68 + t37 + t35;
P(3,1) = t9 - 2 * m1_12 * m1_3 + t72 + t73 + t74 + t12 - 2 * m1_11 * m1_2 - 2 * m1_10 * m1_1 + t11;
P(3,2) = 2 * t89;
P(3,3) = -2 * m2_11 * m2_2 + t37 + t92 - 2 * m2_10 * m2_1 + t95 + t29 + t96 - 2 * m2_12 * m2_3 + t35;
P(4,1) = -2 * m1_9 * m1_6 + t8 + t10 + t7 - 2 * m1_7 * m1_4 + t44 + t45 + t46 - 2 * m1_8 * m1_5;
P(4,2) = 2 * t113;
P(4,3) = -2 * m2_9 * m2_6 + t68 + t64 - 2 * m2_7 * m2_4 - 2 * m2_8 * m2_5 + t34 + t36 + t38 + t65;
P(5,1) = t73 + t8 + t10 + t7 - 2 * m1_10 * m1_4 - 2 * m1_11 * m1_5 - 2 * m1_12 * m1_6 + t74 + t72;
P(5,2) = 2 * t134;
P(5,3) = -2 * m2_12 * m2_6 + t96 + t92 - 2 * m2_10 * m2_4 - 2 * m2_11 * m2_5 + t34 + t36 + t38 + t95;
P(6,1) = t46 + t73 + t44 + t45 + t74 + t72 - 2 * m1_11 * m1_8 - 2 * m1_10 * m1_7 - 2 * m1_12 * m1_9;
P(6,2) = 2 * t155;
P(6,3) = t65 - 2 * m2_10 * m2_7 + t96 + t92 + t95 - 2 * m2_11 * m2_8 + t68 + t64 - 2 * m2_12 * m2_9;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


function dsq=define_distances_btw_control_points()


c1=[1,0,0];
c2=[0,1,0];
c3=[0,0,1];
c4=[0,0,0];

d12=(c1(1)-c2(1))^2 + (c1(2)-c2(2))^2 + (c1(3)-c2(3))^2;
d13=(c1(1)-c3(1))^2 + (c1(2)-c3(2))^2 + (c1(3)-c3(3))^2;
d14=(c1(1)-c4(1))^2 + (c1(2)-c4(2))^2 + (c1(3)-c4(3))^2;
d23=(c2(1)-c3(1))^2 + (c2(2)-c3(2))^2 + (c2(3)-c3(3))^2;
d24=(c2(1)-c4(1))^2 + (c2(2)-c4(2))^2 + (c2(3)-c4(3))^2;
d34=(c3(1)-c4(1))^2 + (c3(2)-c4(2))^2 + (c3(3)-c4(3))^2;

dsq=[d12,d13,d14,d23,d24,d34]';


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function P=compute_constraint_distance_3param_6eq_6unk(m1,m2,m3)

% Copyright (C) <2007>  <Francesc Moreno-Noguer, Vincent Lepetit, Pascal Fua>
% 
% This program is free software: you can redistribute it and/or modify
% it under the terms of the version 3 of the GNU General Public License
% as published by the Free Software Foundation.
% 
% This program is distributed in the hope that it will be useful, but
% WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
% General Public License for more details.       
% You should have received a copy of the GNU General Public License
% along with this program. If not, see <http://www.gnu.org/licenses/>.
%
% Francesc Moreno-Noguer, CVLab-EPFL, September 2007.
% fmorenoguer@gmail.com, http://cvlab.epfl.ch/~fmoreno/ 

%redefine variables name, for compatibility with maple
m1_1=m1(1); 
m1_2=m1(2); 
m1_3=m1(3); 
m1_4=m1(4); 
m1_5=m1(5); 
m1_6=m1(6);
m1_7=m1(7); 
m1_8=m1(8); 
m1_9=m1(9); 
m1_10=m1(10); 
m1_11=m1(11); 
m1_12=m1(12);

m2_1=m2(1); 
m2_2=m2(2); 
m2_3=m2(3); 
m2_4=m2(4); 
m2_5=m2(5); 
m2_6=m2(6);
m2_7=m2(7); 
m2_8=m2(8); 
m2_9=m2(9); 
m2_10=m2(10); 
m2_11=m2(11); 
m2_12=m2(12);

m3_1=m3(1); 
m3_2=m3(2); 
m3_3=m3(3); 
m3_4=m3(4); 
m3_5=m3(5); 
m3_6=m3(6);
m3_7=m3(7); 
m3_8=m3(8); 
m3_9=m3(9); 
m3_10=m3(10); 
m3_11=m3(11); 
m3_12=m3(12);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

t1 = (m1_2 ^ 2);
t4 = (m1_6 ^ 2);
t5 = (m1_3 ^ 2);
t6 = (m1_5 ^ 2);
t11 = (m1_4 ^ 2);
t12 = (m1_1 ^ 2);
t20 = m1_4 * m2_4;
t21 = m1_3 * m2_3;
t22 = m1_5 * m2_5;
t23 = m1_2 * m2_2;
t24 = m1_6 * m2_6;
t25 = m1_1 * m2_1;
t26 = (-m2_4 * m1_1 - m2_5 * m1_2 - m2_6 * m1_3 - m1_6 * m2_3 - m1_4 * m2_1 - m1_5 * m2_2 + t20 + t21 + t22 + t23 + t24 + t25);
t27 = m1_6 * m3_6;
t29 = m1_5 * m3_5;
t30 = m1_4 * m3_4;
t33 = m1_3 * m3_3;
t35 = m1_1 * m3_1;
t38 = m1_2 * m3_2;
t39 = (t27 - m1_6 * m3_3 + t29 + t30 - m1_4 * m3_1 - m3_6 * m1_3 + t33 - m3_5 * m1_2 + t35 - m3_4 * m1_1 - m1_5 * m3_2 + t38);
t40 = (m2_4 ^ 2);
t41 = (m2_2 ^ 2);
t42 = (m2_5 ^ 2);
t43 = (m2_1 ^ 2);
t44 = (m2_6 ^ 2);
t45 = (m2_3 ^ 2);
t53 = m2_4 * m3_4;
t56 = m2_5 * m3_5;
t57 = m2_2 * m3_2;
t60 = m2_1 * m3_1;
t62 = m2_6 * m3_6;
t63 = m2_3 * m3_3;
t65 = (t53 - m2_4 * m3_1 - m3_4 * m2_1 + t56 + t57 - m2_5 * m3_2 - m2_6 * m3_3 + t60 - m3_6 * m2_3 + t62 + t63 - m3_5 * m2_2);
t66 = (m3_5 ^ 2);
t69 = (m3_4 ^ 2);
t70 = (m3_3 ^ 2);
t71 = (m3_2 ^ 2);
t72 = (m3_6 ^ 2);
t75 = (m3_1 ^ 2);
t81 = (m1_9 ^ 2);
t82 = (m1_8 ^ 2);
t83 = (m1_7 ^ 2);
t90 = m1_8 * m2_8;
t91 = m1_7 * m2_7;
t93 = m1_9 * m2_9;
t98 = (-m1_9 * m2_3 + t21 + t90 + t23 + t91 + t25 - m2_7 * m1_1 + t93 - m2_9 * m1_3 - m1_8 * m2_2 - m1_7 * m2_1 - m2_8 * m1_2);
t101 = m1_9 * m3_9;
t102 = m1_7 * m3_7;
t106 = m1_8 * m3_8;
t108 = (-m3_8 * m1_2 + t38 + t33 - m3_9 * m1_3 + t101 + t102 - m1_7 * m3_1 - m1_9 * m3_3 + t35 - m3_7 * m1_1 + t106 - m1_8 * m3_2);
t113 = (m2_8 ^ 2);
t116 = (m2_9 ^ 2);
t117 = (m2_7 ^ 2);
t119 = m2_8 * m3_8;
t122 = m2_9 * m3_9;
t125 = m2_7 * m3_7;
t128 = (t119 - m2_7 * m3_1 - m3_8 * m2_2 + t57 + t63 + t60 + t122 - m2_8 * m3_2 - m3_7 * m2_1 + t125 - m3_9 * m2_3 - m2_9 * m3_3);
t129 = (m3_7 ^ 2);
t130 = (m3_8 ^ 2);
t133 = (m3_9 ^ 2);
t141 = (m1_10 ^ 2);
t142 = (m1_11 ^ 2);
t143 = (m1_12 ^ 2);
t151 = m1_10 * m2_10;
t152 = m1_12 * m2_12;
t154 = m1_11 * m2_11;
t158 = (-m2_10 * m1_1 - m2_12 * m1_3 + t151 + t23 + t25 + t152 + t21 - m1_10 * m2_1 + t154 - m1_11 * m2_2 - m2_11 * m1_2 - m1_12 * m2_3);
t160 = m1_12 * m3_12;
t164 = m1_10 * m3_10;
t165 = m1_11 * m3_11;
t168 = (-m3_10 * m1_1 + t160 - m3_11 * m1_2 + t38 + t33 - m1_12 * m3_3 - m3_12 * m1_3 + t164 + t35 + t165 - m1_11 * m3_2 - m1_10 * m3_1);
t169 = (m2_12 ^ 2);
t170 = (m2_10 ^ 2);
t171 = (m2_11 ^ 2);
t179 = m2_10 * m3_10;
t181 = m2_12 * m3_12;
t185 = m2_11 * m3_11;
t188 = (t57 + t60 + t179 - m2_10 * m3_1 + t181 - m2_12 * m3_3 - m3_12 * m2_3 - m3_10 * m2_1 + t63 + t185 - m2_11 * m3_2 - m3_11 * m2_2);
t191 = (m3_12 ^ 2);
t192 = (m3_10 ^ 2);
t193 = (m3_11 ^ 2);
t212 = (t22 + t91 + t93 - m1_7 * m2_4 + t20 - m2_7 * m1_4 + t90 - m2_8 * m1_5 - m1_8 * m2_5 - m1_9 * m2_6 + t24 - m2_9 * m1_6);
t219 = (t101 + t30 - m3_9 * m1_6 - m3_8 * m1_5 - m1_8 * m3_5 - m3_7 * m1_4 + t27 - m1_9 * m3_6 - m1_7 * m3_4 + t102 + t106 + t29);
t233 = (t53 - m2_8 * m3_5 - m2_9 * m3_6 - m2_7 * m3_4 + t125 + t122 - m3_8 * m2_5 + t62 + t119 + t56 - m3_7 * m2_4 - m3_9 * m2_6);
t254 = (-m2_11 * m1_5 + t154 + t151 - m1_12 * m2_6 + t20 - m1_10 * m2_4 - m2_12 * m1_6 - m2_10 * m1_4 - m1_11 * m2_5 + t152 + t24 + t22);
t261 = (t30 - m3_12 * m1_6 - m1_10 * m3_4 - m1_11 * m3_5 + t160 + t27 + t164 + t165 - m3_11 * m1_5 - m3_10 * m1_4 - m1_12 * m3_6 + t29);
t275 = (-m3_10 * m2_4 + t56 - m2_10 * m3_4 + t62 - m3_12 * m2_6 - m2_11 * m3_5 + t53 - m3_11 * m2_5 - m2_12 * m3_6 + t179 + t181 + t185);
t296 = (-m2_12 * m1_9 + t152 + t91 + t90 - m1_11 * m2_8 + t151 - m2_11 * m1_8 + t93 - m1_10 * m2_7 - m1_12 * m2_9 - m2_10 * m1_7 + t154);
t303 = (t164 + t102 - m3_10 * m1_7 + t160 - m3_12 * m1_9 - m1_10 * m3_7 - m3_11 * m1_8 + t101 - m1_12 * m3_9 + t106 + t165 - m1_11 * m3_8);
t317 = (t125 + t119 - m3_12 * m2_9 - m3_10 * m2_7 - m2_11 * m3_8 - m3_11 * m2_8 + t122 + t179 - m2_12 * m3_9 + t181 - m2_10 * m3_7 + t185);
P(1,1) = t1 - 2 * m1_4 * m1_1 + t4 + t5 + t6 - 2 * m1_5 * m1_2 - 2 * m1_6 * m1_3 + t11 + t12;
P(1,2) = 2 * t26;
P(1,3) = 2 * t39;
P(1,4) = t40 + t41 + t42 + t43 + t44 + t45 - 2 * m2_5 * m2_2 - 2 * m2_6 * m2_3 - 2 * m2_4 * m2_1;
P(1,5) = 2 * t65;
P(1,6) = t66 - 2 * m3_4 * m3_1 + t69 + t70 + t71 + t72 - 2 * m3_6 * m3_3 + t75 - 2 * m3_5 * m3_2;
P(2,1) = -2 * m1_8 * m1_2 + t12 + t81 + t5 + t82 + t83 + t1 - 2 * m1_9 * m1_3 - 2 * m1_7 * m1_1;
P(2,2) = 2 * t98;
P(2,3) = 2 * t108;
P(2,4) = -2 * m2_8 * m2_2 - 2 * m2_9 * m2_3 + t113 - 2 * m2_7 * m2_1 + t116 + t117 + t41 + t43 + t45;
P(2,5) = 2 * t128;
P(2,6) = t75 + t70 + t129 + t71 + t130 - 2 * m3_9 * m3_3 + t133 - 2 * m3_8 * m3_2 - 2 * m3_7 * m3_1;
P(3,1) = -2 * m1_11 * m1_2 + t141 + t142 + t12 + t1 + t5 + t143 - 2 * m1_10 * m1_1 - 2 * m1_12 * m1_3;
P(3,2) = 2 * t158;
P(3,3) = 2 * t168;
P(3,4) = t169 + t41 + t43 + t45 + t170 + t171 - 2 * m2_12 * m2_3 - 2 * m2_10 * m2_1 - 2 * m2_11 * m2_2;
P(3,5) = 2 * t188;
P(3,6) = t71 - 2 * m3_12 * m3_3 + t75 + t191 + t70 + t192 + t193 - 2 * m3_10 * m3_1 - 2 * m3_11 * m3_2;
P(4,1) = -2 * m1_9 * m1_6 + t11 + t4 + t81 + t82 + t6 + t83 - 2 * m1_7 * m1_4 - 2 * m1_8 * m1_5;
P(4,2) = 2 * t212;
P(4,3) = 2 * t219;
P(4,4) = t117 + t113 - 2 * m2_8 * m2_5 - 2 * m2_7 * m2_4 - 2 * m2_9 * m2_6 + t116 + t40 + t42 + t44;
P(4,5) = 2 * t233;
P(4,6) = t129 - 2 * m3_9 * m3_6 + t69 + t66 + t72 + t133 - 2 * m3_8 * m3_5 - 2 * m3_7 * m3_4 + t130;
P(5,1) = t4 + t143 + t11 - 2 * m1_12 * m1_6 - 2 * m1_11 * m1_5 - 2 * m1_10 * m1_4 + t6 + t141 + t142;
P(5,2) = 2 * t254;
P(5,3) = 2 * t261;
P(5,4) = t170 + t171 + t169 - 2 * m2_10 * m2_4 - 2 * m2_11 * m2_5 - 2 * m2_12 * m2_6 + t40 + t42 + t44;
P(5,5) = 2 * t275;
P(5,6) = t69 + t66 + t72 - 2 * m3_12 * m3_6 - 2 * m3_10 * m3_4 + t193 - 2 * m3_11 * m3_5 + t192 + t191;
P(6,1) = t142 - 2 * m1_10 * m1_7 + t141 + t83 + t143 + t82 - 2 * m1_12 * m1_9 + t81 - 2 * m1_11 * m1_8;
P(6,2) = 2 * t296;
P(6,3) = 2 * t303;
P(6,4) = t171 + t170 - 2 * m2_12 * m2_9 - 2 * m2_10 * m2_7 - 2 * m2_11 * m2_8 + t113 + t169 + t116 + t117;
P(6,5) = 2 * t317;
P(6,6) = -2 * m3_11 * m3_8 + t193 + t133 + t129 + t192 + t130 - 2 * m3_12 * m3_9 + t191 - 2 * m3_10 * m3_7;



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


function [Xc_opt,R_opt,T_opt,err_opt,iter]=optimize_betas_gauss_newton(Km,Cw,Beta0,Alph,Xw,U,A)



n=size(Beta0,2);

[Beta_opt,err,iter]=gauss_newton(Km,Cw,Beta0);

%Extract control point camera coordinates from Betas and Kernel
X=zeros(12,1);
for i=1:n
   X=X+Beta_opt(i)*Km(:,i); 
end

Cc=zeros(4,3);
for i=1:4
    Cc(i,:)=X(3*i-2:3*i);
end


%check sign of the determinant (keep orienation of the control points)
s_Cw=sign_determinant(Cw);
s_Cc=sign_determinant(Cc);
Cc=Cc*(s_Cw/s_Cc);

%Reconstruct and compute error=
Xc_opt=Alph*Cc; %reconstruction: points in camera coordinate system
[R_opt,T_opt]=getrotT(Xw,Xc_opt);  %solve exterior orientation

[err_opt,Urep_opt]=reprojection_error_usingRT(Xw,U,R_opt,T_opt,A);



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


function [Beta_opt,err,iter]=gauss_newton(K,Cw,Beta0)



L=compute_L6_10(K);
rho=compute_rho(Cw);

current_betas=Beta0;
  
n_iterations=5; %max number of iterations. Usually 4-5 is enought.
for k=1:n_iterations
    [A,b]=compute_A_and_b_Gauss_Newton(current_betas,rho,L);
    dbeta=inv(A'*A)*A'*b;
    current_betas=current_betas+dbeta';
    error(k)=b'*b;
    iter(k).betas=current_betas;
    iter(k).error=error(k);    
end

Beta_opt=current_betas;
err=error(end);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


function [A,b]=compute_A_and_b_Gauss_Newton(current_betas,rho,L)





A=zeros(6,4);
b=zeros(6,1);

cb=current_betas;

B=[cb(1)*cb(1),
   cb(1)*cb(2),
   cb(2)*cb(2),
   cb(1)*cb(3),
   cb(2)*cb(3),
   cb(3)*cb(3),
   cb(1)*cb(4),
   cb(2)*cb(4),
   cb(3)*cb(4),
   cb(4)*cb(4)];

for i=1:6
    A(i,1)=2*cb(1)*L(i,1)+cb(2)*L(i,2)+cb(3)*L(i,4)+cb(4)*L(i,7);
    A(i,2)=cb(1)*L(i,2)+2*cb(2)*L(i,3)+cb(3)*L(i,5)+cb(4)*L(i,8);
    A(i,3)=cb(1)*L(i,4)+cb(2)*L(i,5)+2*cb(3)*L(i,6)+cb(4)*L(i,9);
    A(i,4)=cb(1)*L(i,7)+cb(2)*L(i,8)+cb(3)*L(i,9)+2*cb(4)*L(i,10);
        
    b(i)=(rho(i)-L(i,:)*B);    
end
    
 %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function L=compute_L6_10(K)

% COMPUTE_L6_10  
%
% Copyright (C) <2007>  <Francesc Moreno-Noguer, Vincent Lepetit, Pascal Fua>
% 
% This program is free software: you can redistribute it and/or modify
% it under the terms of the version 3 of the GNU General Public License
% as published by the Free Software Foundation.
% 
% This program is distributed in the hope that it will be useful, but
% WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
% General Public License for more details.       
% You should have received a copy of the GNU General Public License
% along with this program. If not, see <http://www.gnu.org/licenses/>.
%
% Francesc Moreno-Noguer, CVLab-EPFL, October 2007.
% fmorenoguer@gmail.com, http://cvlab.epfl.ch/~fmoreno/ 

L=zeros(6,10);

%extract vectors
v1=K(:,1);
v2=K(:,2);
v3=K(:,3);
v4=K(:,4);

%compute differences
dx112 = v1(1) - v1(4);  dy112 = v1(2) - v1(5);  dz112 = v1(3) - v1(6);
dx113 = v1(1) - v1(7);  dy113 = v1(2) - v1(8);  dz113 = v1(3) - v1(9);
dx114 = v1(1) - v1(10); dy114 = v1(2) - v1(11); dz114 = v1(3) - v1(12);
dx123 = v1(4) - v1(7);  dy123 = v1(5) - v1(8);  dz123 = v1(6) - v1(9);
dx124 = v1(4) - v1(10); dy124 = v1(5) - v1(11); dz124 = v1(6) - v1(12);
dx134 = v1(7) - v1(10); dy134 = v1(8) - v1(11); dz134 = v1(9) - v1(12);

dx212 = v2(1) - v2(4);  dy212 = v2(2) - v2(5);  dz212 = v2(3) - v2(6);
dx213 = v2(1) - v2(7);  dy213 = v2(2) - v2(8);  dz213 = v2(3) - v2(9);
dx214 = v2(1) - v2(10); dy214 = v2(2) - v2(11); dz214 = v2(3) - v2(12);
dx223 = v2(4) - v2(7);  dy223 = v2(5) - v2(8);  dz223 = v2(6) - v2(9);
dx224 = v2(4) - v2(10); dy224 = v2(5) - v2(11); dz224 = v2(6) - v2(12);
dx234 = v2(7) - v2(10); dy234 = v2(8) - v2(11); dz234 = v2(9) - v2(12);

dx312 = v3(1) - v3(4);  dy312 = v3(2) - v3(5);  dz312 = v3(3) - v3(6);
dx313 = v3(1) - v3(7);  dy313 = v3(2) - v3(8);  dz313 = v3(3) - v3(9);
dx314 = v3(1) - v3(10); dy314 = v3(2) - v3(11); dz314 = v3(3) - v3(12);
dx323 = v3(4) - v3(7);  dy323 = v3(5) - v3(8);  dz323 = v3(6) - v3(9);
dx324 = v3(4) - v3(10); dy324 = v3(5) - v3(11); dz324 = v3(6) - v3(12);
dx334 = v3(7) - v3(10); dy334 = v3(8) - v3(11); dz334 = v3(9) - v3(12);

dx412 = v4(1) - v4(4);  dy412 = v4(2) - v4(5);  dz412 = v4(3) - v4(6);
dx413 = v4(1) - v4(7);  dy413 = v4(2) - v4(8);  dz413 = v4(3) - v4(9);
dx414 = v4(1) - v4(10); dy414 = v4(2) - v4(11); dz414 = v4(3) - v4(12);
dx423 = v4(4) - v4(7);  dy423 = v4(5) - v4(8);  dz423 = v4(6) - v4(9);
dx424 = v4(4) - v4(10); dy424 = v4(5) - v4(11); dz424 = v4(6) - v4(12);
dx434 = v4(7) - v4(10); dy434 = v4(8) - v4(11); dz434 = v4(9) - v4(12);


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
L(1,1) =        dx112 * dx112 + dy112 * dy112 + dz112 * dz112;      %b1*b1
L(1,2) = 2.0 *  (dx112 * dx212 + dy112 * dy212 + dz112 * dz212);    %b1*b2
L(1,3) =        dx212 * dx212 + dy212 * dy212 + dz212 * dz212;      %b2*b2
L(1,4) = 2.0 *  (dx112 * dx312 + dy112 * dy312 + dz112 * dz312);    %b1*b3
L(1,5) = 2.0 *  (dx212 * dx312 + dy212 * dy312 + dz212 * dz312);    %b2*b3
L(1,6) =        dx312 * dx312 + dy312 * dy312 + dz312 * dz312;      %b3*b3
L(1,7) = 2.0 *  (dx112 * dx412 + dy112 * dy412 + dz112 * dz412);    %b1*b4
L(1,8) = 2.0 *  (dx212 * dx412 + dy212 * dy412 + dz212 * dz412);    %b2*b4
L(1,9) = 2.0 *  (dx312 * dx412 + dy312 * dy412 + dz312 * dz412);    %b3*b4
L(1,10) =       dx412 * dx412 + dy412 * dy412 + dz412 * dz412;      %b4*b4


L(2,1) =        dx113 * dx113 + dy113 * dy113 + dz113 * dz113;
L(2,2) = 2.0 *  (dx113 * dx213 + dy113 * dy213 + dz113 * dz213);
L(2,3) =        dx213 * dx213 + dy213 * dy213 + dz213 * dz213;
L(2,4) = 2.0 *  (dx113 * dx313 + dy113 * dy313 + dz113 * dz313);
L(2,5) = 2.0 *  (dx213 * dx313 + dy213 * dy313 + dz213 * dz313);
L(2,6) =        dx313 * dx313 + dy313 * dy313 + dz313 * dz313;
L(2,7) = 2.0 *  (dx113 * dx413 + dy113 * dy413 + dz113 * dz413);
L(2,8) = 2.0 *  (dx213 * dx413 + dy213 * dy413 + dz213 * dz413);
L(2,9) = 2.0 *  (dx313 * dx413 + dy313 * dy413 + dz313 * dz413);
L(2,10) =       dx413 * dx413 + dy413 * dy413 + dz413 * dz413;


L(3,1) =        dx114 * dx114 + dy114 * dy114 + dz114 * dz114;
L(3,2) = 2.0 *  (dx114 * dx214 + dy114 * dy214 + dz114 * dz214);
L(3,3) =        dx214 * dx214 + dy214 * dy214 + dz214 * dz214;
L(3,4) = 2.0 *  (dx114 * dx314 + dy114 * dy314 + dz114 * dz314);
L(3,5) = 2.0 *  (dx214 * dx314 + dy214 * dy314 + dz214 * dz314);
L(3,6) =        dx314 * dx314 + dy314 * dy314 + dz314 * dz314;
L(3,7) = 2.0 *  (dx114 * dx414 + dy114 * dy414 + dz114 * dz414);
L(3,8) = 2.0 *  (dx214 * dx414 + dy214 * dy414 + dz214 * dz414);
L(3,9) = 2.0 *  (dx314 * dx414 + dy314 * dy414 + dz314 * dz414);
L(3,10) =       dx414 * dx414 + dy414 * dy414 + dz414 * dz414;


L(4,1) =        dx123 * dx123 + dy123 * dy123 + dz123 * dz123;
L(4,2) = 2.0 *  (dx123 * dx223 + dy123 * dy223 + dz123 * dz223);
L(4,3) =        dx223 * dx223 + dy223 * dy223 + dz223 * dz223;
L(4,4) = 2.0 *  (dx123 * dx323 + dy123 * dy323 + dz123 * dz323);
L(4,5) = 2.0 *  (dx223 * dx323 + dy223 * dy323 + dz223 * dz323);
L(4,6) =        dx323 * dx323 + dy323 * dy323 + dz323 * dz323;
L(4,7) = 2.0 *  (dx123 * dx423 + dy123 * dy423 + dz123 * dz423);
L(4,8) = 2.0 *  (dx223 * dx423 + dy223 * dy423 + dz223 * dz423);
L(4,9) = 2.0 *  (dx323 * dx423 + dy323 * dy423 + dz323 * dz423);
L(4,10) =       dx423 * dx423 + dy423 * dy423 + dz423 * dz423;


L(5,1) =        dx124 * dx124 + dy124 * dy124 + dz124 * dz124;
L(5,2) = 2.0 *  (dx124 * dx224 + dy124 * dy224 + dz124 * dz224);
L(5,3) =        dx224 * dx224 + dy224 * dy224 + dz224 * dz224;
L(5,4) = 2.0 * ( dx124 * dx324 + dy124 * dy324 + dz124 * dz324);
L(5,5) = 2.0 * (dx224 * dx324 + dy224 * dy324 + dz224 * dz324);
L(5,6) =        dx324 * dx324 + dy324 * dy324 + dz324 * dz324;
L(5,7) = 2.0 * ( dx124 * dx424 + dy124 * dy424 + dz124 * dz424);
L(5,8) = 2.0 * ( dx224 * dx424 + dy224 * dy424 + dz224 * dz424);
L(5,9) = 2.0 * ( dx324 * dx424 + dy324 * dy424 + dz324 * dz424);
L(5,10) =       dx424 * dx424 + dy424 * dy424 + dz424 * dz424;


L(6,1) =        dx134 * dx134 + dy134 * dy134 + dz134 * dz134;
L(6,2) = 2.0 * ( dx134 * dx234 + dy134 * dy234 + dz134 * dz234);
L(6,3) =        dx234 * dx234 + dy234 * dy234 + dz234 * dz234;
L(6,4) = 2.0 * ( dx134 * dx334 + dy134 * dy334 + dz134 * dz334);
L(6,5) = 2.0 * ( dx234 * dx334 + dy234 * dy334 + dz234 * dz334);
L(6,6) =        dx334 * dx334 + dy334 * dy334 + dz334 * dz334;
L(6,7) = 2.0 *  (dx134 * dx434 + dy134 * dy434 + dz134 * dz434);
L(6,8) = 2.0 *  (dx234 * dx434 + dy234 * dy434 + dz234 * dz434);
L(6,9) = 2.0 *  (dx334 * dx434 + dy334 * dy434 + dz334 * dz434);
L(6,10)=       dx434 * dx434 + dy434 * dy434 + dz434 * dz434;


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function rho=compute_rho(Cw)

% COMPUTE_RHO
c0=Cw(1,:);
c1=Cw(2,:);
c2=Cw(3,:);
c3=Cw(4,:);

rho(1) = dist2(c0,c1);
rho(2) = dist2(c0,c2);
rho(3) = dist2(c0,c3);
rho(4) = dist2(c1,c2);
rho(5) = dist2(c1,c3);
rho(6) = dist2(c2,c3);
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function s=sign_determinant(C)

% SIGN_DETERMINANT
c0=C(4,:)';
c1=C(1,:)';
c2=C(2,:)';
c3=C(3,:)';


v1=c1-c0;
v2=c2-c0;
v3=c3-c0;

M=[v1,v2,v3];
detM=det(M);
s=sign(detM);


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function dist=dist2(v1,v2)



diff=v1-v2;
dist=sum(diff.^2);