% estimate_minimum_reprojection_error2 
%
%       Francesc Moreno-Noguer, Vincent Lepetit, Pascal Fua
%       Pose Priors for Simultaneously Solving Alignment and Correspondence
%       European Conference on Computer Vision (ECCV), 2008
%
% Copyright (C) <2009>  <Francesc Moreno-Noguer, Vincent Lepetit, Pascal Fua>
% 
% This program is free software: you can redistribute it and/or modify
% it under the terms of the version 3 of the GNU General Public License
% as published by the Free Software Foundation.
% 
% This program is distributed in the hope that it will be useful, but
% WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
% General Public License for more details.       
% You should have received a copy of the GNU General Public License
% along with this program. If not, see <http://www.gnu.org/licenses/>.
%
% Francesc Moreno-Noguer, Institut de Robtica i Informtica Industrial (CSIC/UPC), May 2009.
% fmoreno@iri.upc.edu, http://www-iri.upc.es/people/fmoreno/


function [err_min,percentage_outliers,Idx_matching3d_2d,Idx_outliers]=estimate_minimum_reprojection_error2(U,Uproj,tree,node,SENSOR_NOISE)

M=size(Uproj,1);
N=size(U,1);
original_idx3d=[1:M];
original_idx2d=[1:N];

%% 1.-First: remove all the matches already done in the tree
ancestors=tree.ancestors(node,:);
ancestors=ancestors(ancestors~=0);

match=tree.match(ancestors,:);
match(end+1,:)=tree.match(node,:); %we add the match of the last node
match_3d=match(:,1);
match_2d=match(:,2);

%remove the first node (with match (0,0));
match_3d=match_3d(2:end);
match_2d=match_2d(2:end);

%detect initial outliers: with match_2d==0
out=find(match_2d==0);
in=find(match_2d~=0);
match_2d_noout=match_2d;
match_2d_noout(out,:)=[];

%Remove all assignments already done from U and Uproj
Uproj2=Uproj;
Uproj2(match_3d,:)=[];

U2=U;
U2(match_2d_noout,:)=[];

%Generate inliers and outliers list of matchings
Inliers3d=Uproj(match_3d(in),:);
Inliers2d=U(match_2d(in),:);
Outliers3d=Uproj(match_3d(out),:);

D1=sqrt((Inliers3d(:,1)-Inliers2d(:,1)).^2+(Inliers3d(:,2)-Inliers2d(:,2)).^2);
D2=SENSOR_NOISE*ones(1,size(Outliers3d,1));
Minimum_cost0=[D1',D2];

%List of correspondences
Idx_matching3d_2d=[match_3d(in),match_2d(in)];
Idx_ouliers3d=[match_3d(out)];
Idx_matching3d_2d=[Idx_matching3d_2d;[Idx_ouliers3d,zeros(size(Idx_ouliers3d))]];
original_idx3d(match_3d)=[];
original_idx2d(match_2d(match_2d~=0))=[];


%% 2.-Second: make the correspondence of the easy points
M2=size(Uproj2,1);
N2=size(U2,1);

Uproj_big_x=repmat(Uproj2(:,1),[1,N2]);
Uproj_big_y=repmat(Uproj2(:,2),[1,N2]);

U_big_x=repmat(U2(:,1)',[M2,1]);
U_big_y=repmat(U2(:,2)',[M2,1]);

Dx=(Uproj_big_x-U_big_x).^2;
Dy=(Uproj_big_y-U_big_y).^2;
D=sqrt(Dx+Dy);
Dini=D;

%find candidates 2d which are selected by only one point 3d (these are
%unique matchings)
if M2>0
    for n=1:M2
        points_2d=D(n,:);
        [min_dist(n),idx]=min(points_2d);
        candidate(n)=idx;
    end
    [idxnonrepeated,idxrepeated,nrep,allidxrepeated]=nonrepeated(candidate);
    nunique_matches=size(idxnonrepeated,2);

    Minimum_cost1=[];
    if nunique_matches>0
        cm=[idxnonrepeated',candidate(idxnonrepeated)']; %correctmatching indices
        ncm=size(cm,1);
        on=ones(nunique_matches,1);
        linear_indices=(cm(:,2)-on)*M2+cm(:,1);
        Minimum_cost1=D(linear_indices)';

        %remove the rows and colums of D corresponding to these points
        D(cm(:,1),:)=[];
        D(:,cm(:,2))=[];

        %Update original list of correspondences
        Idx_matching3d_2d_new=[original_idx3d(cm(:,1))',original_idx2d(cm(:,2))'];
        Idx_matching3d_2d=[Idx_matching3d_2d;Idx_matching3d_2d_new];
        original_idx3d(cm(:,1))=[];
        original_idx2d(cm(:,2))=[];
    end
    
else
    Minimum_cost1=[];
end


%solve points with various candidates
%I will first assign the point with minimum distance
[M3,N3]=size(D);
if M3>0
    Minimum_cost2=[];
    for i=1:M3
        [m,minpos]=min(D(:));
        %row and column of minpos. Remove row and column
        res=mod(minpos,M3);
        if res==0
            col=minpos/M3;
        else
            col=(minpos-res)/M3+1;
        end
        row=minpos-(col-1)*M3;
        D(row,:)=[];
        D(:,col)=[];

        Minimum_cost2=[Minimum_cost2,m];
        M3=M3-1;
        N3=N3-1;
        
        %Update original list of correspondences
        Idx_matching3d_2d_new=[original_idx3d(row),original_idx2d(col)];
        Idx_matching3d_2d=[Idx_matching3d_2d;Idx_matching3d_2d_new];
        original_idx3d(row)=[];
        original_idx2d(col)=[];
    end
else
    Minimum_cost2=[];
end
Minimum_cost=[Minimum_cost0,Minimum_cost1,Minimum_cost2];

   
%% 3.-Determine outliers  
%set distances larger than sensor noise to outliers
outliers=find(Minimum_cost>=SENSOR_NOISE);
inliers=find(Minimum_cost<SENSOR_NOISE);
Minimum_cost(outliers)=SENSOR_NOISE;

err_min=mean(Minimum_cost);
num_outliers=size(outliers,2);
percentage_outliers=num_outliers/M;
num_inliers=M-num_outliers;

idx3d=Idx_matching3d_2d(:,1);
Idx_outliers=idx3d(outliers);

Idx_matching3d_2d=Idx_matching3d_2d(inliers,:);


%% Plot results
plot_results=0;
if plot_results

    o=find(Dini>SENSOR_NOISE);
    Dini(o)=SENSOR_NOISE;
    
    figure; 
    imagesc(Dini); colormap gray;
    %imshow(Dini/max(Dini(:)));
    xlabel('N image points');
    ylabel('M model points');

    figure;
    plot(Uproj(:,1),Uproj(:,2),'.','color','r');  hold on;
    plot(U(:,1),U(:,2),'.');

    npoints_proj=size(Uproj,1);
    for i=1:npoints_proj
        txt=sprintf('%d',i);
        text(Uproj(i,1),Uproj(i,2),txt,'fontsize',15);
    end

    npoints_img=size(U,1);
    for i=1:npoints_img
        txt=sprintf('%d',i);
        text(U(i,1),U(i,2),txt);
    end
end