% mainBPnP 
%
%       Francesc Moreno-Noguer, Vincent Lepetit, Pascal Fua
%       Pose Priors for Simultaneously Solving Alignment and Correspondence
%       European Conference on Computer Vision (ECCV), 2008
%
% Copyright (C) <2009>  <Francesc Moreno-Noguer, Vincent Lepetit, Pascal Fua>
% 
% This program is free software: you can redistribute it and/or modify
% it under the terms of the version 3 of the GNU General Public License
% as published by the Free Software Foundation.
% 
% This program is distributed in the hope that it will be useful, but
% WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
% General Public License for more details.       
% You should have received a copy of the GNU General Public License
% along with this program. If not, see <http://www.gnu.org/licenses/>.
%
% Francesc Moreno-Noguer, Institut de Robtica i Informtica Industrial (CSIC/UPC), May 2009.
% fmoreno@iri.upc.edu, http://www-iri.upc.es/people/fmoreno/


clear all; close all;


%% Paths
addpath BPose; addpath General_functions; addpath Mixture_of_gaussians; 
path1_='Results/Pose_Priors/';
path2_='Results/InputData/';
path3_='Results/BP/';

%% Parameters to define the experiments 
Ns=[30];%[25:5:80]; %number of model points
PCs=0.4;%[0.2:0.2:0.8]; %[0.2:0.2:0.8]; %percentage of clutter
PDs=0.8;%[0.4:0.2:0.8];%[0.2:0.2:0.8]; %percentage of detection 
NUMBER_EXPERIMENTS=1; %50; %number of experiments for each combination of parameters
std_noise=2; %noise in the measurements (in pixels)

%% Parameters for the BPnP algorithm
num_poses=20; %number of gaussians for the EM algorithm (number of pose priors)
SENSOR_NOISE=10;
MIN_ERROR_THRESHOLD=6.0; %minimum reprojection error (in pixels) to exit.
KICKOUT_THRESHOLD=0.2;  %used to discard the current pose prior when a large number of consecutive model points are not detected
THRESHOLD_MAH=2; THRESHOLD_EUCL=SENSOR_NOISE; %define the search region for each projected point

%% Iterate for all the experimental parameters
for n=Ns
    %create directories if do not exist
    txt=sprintf('N%d/',n);
    path1=strcat(path1_,txt); if ~exist(path1) mkdir(path1); end
    path2=strcat(path2_,txt); if ~exist(path2) mkdir(path2); end
    path3=strcat(path3_,txt); if ~exist(path3) mkdir(path3); end

    for pc=PCs
        for pd=PDs
            %1.-Load/Generate 3D model--------------------------------------------------------
            Xw=generate_3D_model(n); %random points

            %2.-Generate Pose Priors---------------------------------------------------------------
            name=return_name3(path1,'SampledPoses',pc,pd,'mat');
            if ~exist(name)
                [P,Prob,x1,x2,Mu,Cov,A]=generate_pose_priors(Xw,num_poses);
                [Mu,Cov,Prob]=sort_poses(Mu,Cov,Prob);
                save(name,'P','Prob','x1','x2','Mu','Cov','A');
            else
                load(name,'P','Prob','x1','x2','Mu','Cov','A');
            end
            
            %Precompute probabilities for cutting branches 
            [Prob_outliers,Num_outliers]=precompute_outlier_probabilities(n,pd);
            TERMINATION_THRESHOLD=max(4,floor(0.8*pd*n)); %Same termination threshold as in softposit algorithm. Related to the number of inliers.

            for exp=1:NUMBER_EXPERIMENTS
                %3.-Generate 2D input data---------------------------------------------------------
                [U,Ureal,Rtrue,Ttrue,Pose_true,Xw_no_occluded]=generate_2D_noisy_input_data2(Xw,P,A,pc,pd,std_noise);

                %4.-Blind PnP Algorithm starts here -----------------------------------------------
                err=10000;
                num_init_poses=size(Cov,3);
                terminated=0;
                clear t;
                for i=1:num_init_poses
                    %mean and covariances for each pose
                    Sigma_ini=Cov(:,:,i);
                    Pose_ini=Mu(:,i);
                    Sigma_sensor=[SENSOR_NOISE 0; 0 SENSOR_NOISE]; %pixel variance in u and v direction (image plane)

                    init_tree(); %initialization of the tree
                    if i>1
                        tree.min_error_optimal=min(err);
                        tree.error=min(err);
                    end

                    tic;
                    [tree,terminated]=recursive_function(A,Pose_ini,Sigma_ini,Sigma_sensor,SENSOR_NOISE,Xw,U,idx2consider,tree,parent,THRESHOLD_MAH,THRESHOLD_EUCL,Ureal,Prob_outliers,MIN_ERROR_THRESHOLD,TERMINATION_THRESHOLD,terminated,KICKOUT_THRESHOLD);
                    t(i)=toc;
                    if tree.min_error_optimal<min(err)  err(i)=tree.min_error_optimal;
                    else err(i)=10000;  tree.error=10000;
                    end
                    trees(i).tree=tree;
                    if terminated  fprintf('TERMINATED!!!! BlindPose%d, n=%d, pd=%.1f, pc=%.1f,  time=%.1f, Err=%.1f\n',i,n,pd,pc,t(i),err(i)); break;
                    else  fprintf('BlindPose%d, n=%d, pd=%.1f, pc=%.1f,  time=%.1f, Err=%.1f\n',i,n,pd,pc,t(i),err(i));
                    end
                end
                %take the best solution and refine it using a PnP algorithm
                [min_err,min_idx]=min(err);
                Pose_ini=Mu(:,min_idx);
                tree=trees(min_idx).tree;
                if min_err==10000 %to indicate the experiment didn't converge
                    R=ones(3,3)*1000; T=ones(3,1)*10000;
                    err_repr_true=10000;err_repr_est=10000; err_rot=10000; err_trans=10000;
                    percentage_outliers=10000;Idx_matching3d_2d=10000;Idx_outliers=10000;
                else  %refine solution using PnP
                    [R,T,err_repr_true,err_repr_est,err_rot,err_trans,percentage_outliers,Idx_matching3d_2d,Idx_outliers]=error_best_pose(tree,SENSOR_NOISE,A,Xw,U,Pose_ini,Xw_no_occluded,Ureal,Rtrue,Ttrue,0);
                end
                total_time=sum(t);
                results1(exp,:)=[exp,err_repr_est,err_rot,err_trans,percentage_outliers,total_time];
                Poses1(exp).R=R; Poses1(exp).T=T;    

                fprintf('\n------------------------------------\n');
                fprintf('Experiment %d, n=%d, pd=%.1f, pc=%.1f\n',exp,n,pd,pc);
                fprintf('BlindPose: Erepr:%.2f, Erot:%.2f, Etrans:%.2f, Perc.Outliers:%.2f, Time:%.2f\n',results1(exp,2),results1(exp,3),results1(exp,4),results1(exp,5),results1(exp,6));
                fprintf('------------------------------------\n');
                
                %save data for the experiments
                name=return_name2(path2,'InputData',pc,pd,exp,'mat');
                save(name,'A','Xw','Xw_no_occluded','Ureal','Rtrue','Ttrue','Pose_true','U','pc','pd','std_noise');               
            end %NUMBER_EXPERIMENTS

            %save results all experiment
            name=return_name3(path3,'ResultsBP',pc,pd,'mat');
            save(name,'results1','Poses1');

        end %PDs
    end %PCs
end %for n=Ns


%% Plot results
if NUMBER_EXPERIMENTS==1
    %model points + sample camera centers
    figure; hold on;
    plot3(Xw_no_occluded(:,1),Xw_no_occluded(:,2),Xw_no_occluded(:,3),'.','color',[0 0 0],'markersize',10);
    plot3(x2(:,1),x2(:,2),x2(:,3),'.','color',[1 0 0]);
    title('Model Points + Sample Camera Centers');
    axis equal;  view(-26,30);

    %True projection + Projection using estimated pose
    figure; hold on;
    Uest=project_points(A,Poses1.R,Poses1.T,Xw);
    plot(Ureal(:,1),Ureal(:,2),'.','markersize',10,'color',[0 0 0]);
    plot(Uest(:,1),Uest(:,2),'o','markersize',10,'color',[0 1 0],'linewidth',2);    
    plot(Uest(Idx_outliers,1),Uest(Idx_outliers,2),'x','markersize',10,'color',[1 0 0],'linewidth',2);  
    legend('Ground truth projections','Estimated projections','Classified as outliers');
    %xlim([0 640]); ylim([0 480]);
end

